# -*- coding: utf-8 -*-

import os
from s3pathlib import context
from boto_session_manager import BotoSesManager
from simple_aws_ec2.api import Ec2Instance


from .runtime import (
    IS_LOCAL,
    IS_GITHUB_CI,
    IS_EC2,
    IS_CODEBUILD_CI,
    IS_LAMBDA,
    IS_BATCH,
    IS_FARGATE,
)

# environment aware boto session manager
if IS_LOCAL:  # put production first
    bsm = BotoSesManager(
        profile_name="bmt_app_dev_us_east_1",
        region_name="us-east-1",  # hard coded region, because we know what we are doing
    )
elif IS_GITHUB_CI:
    bsm = BotoSesManager(
        region_name="us-east-1",  # hard coded region, because we know what we are doing
    )
elif IS_EC2:
    bsm = BotoSesManager(
        region_name=Ec2Instance.get_placement_region(),
    )
elif IS_CODEBUILD_CI:
    bsm = BotoSesManager(
        region_name=os.environ["AWS_REGION"],
    )
elif IS_LAMBDA:
    bsm = BotoSesManager(
        region_name=os.environ["AWS_DEFAULT_REGION"],
    )
elif IS_BATCH or IS_FARGATE:
    raise NotImplementedError
else:  # pragma: no cover
    raise NotImplementedError

context.attach_boto_session(boto_ses=bsm.boto_ses)



# -*- coding: utf-8 -*-

"""
该模块为使用 ``acore_server_config`` 库的外部项目提供了一套能读取服务器 config 的接口.
相比之下 ``acore_server_config.config.init`` 模块是为 ``acore_server_config`` 库
内部用来从本地文件系统读取配置数据的, 外部项目不应该使用它.

该模块有两个 Public API:

- :class:`Ec2ConfigLoader`: 用于在 EC2 上运行脚本, 用 "自省" 的方式获得自己的配置数据.
- :class:`ConfigLoader`: 用于在任意其他环境显式的加载配置数据.
"""

import dataclasses
import typing as T

from s3pathlib import S3Path
from simple_aws_ec2.api import Ec2Instance
from acore_constants.api import TagKey

from ..boto_ses import bsm as default_bsm

from .define import EnvEnum, Env, Config, Server


if T.TYPE_CHECKING:  # pragma: no cover
    from boto_session_manager import BotoSesManager


def _get_default_s3folder_config(bsm: "BotoSesManager") -> str:
    """
    获得默认的 S3 配置数据的根目录.
    """
    return (
        S3Path(f"s3://{bsm.aws_account_id}-{bsm.aws_region}-artifacts")
        .joinpath(
            "projects",
            "acore_server_config",
            "config",
        )
        .to_dir()
    ).uri


def _get_default_parameter_name_prefix() -> str:
    """
    获得默认的 AWS Parameter Store 的参数名前缀. 这取决于我们 deploy 的时候用的前缀是什么.
    """
    return "acore_server_config"


def get_this_server_id(bsm: "BotoSesManager") -> str:  # pragma: no cover
    """
    在 EC2 上通过 "自省", 获得这个服务器的 server_id. 它的 naming convention 是
    ``${env_name}-${server_name}``.
    """
    ec2_inst = Ec2Instance.from_ec2_inside(bsm.ec2_client)
    server_id = ec2_inst.tags[TagKey.SERVER_ID]
    return server_id


def parse_server_id(server_id: str) -> T.Tuple[str, str]:
    """
    解析 server_id, 返回 (env_name, server_name) 的 tuple.
    """
    env_name, server_name = server_id.split("-", 1)
    return env_name, server_name


def get_config(
    bsm: "BotoSesManager" = default_bsm,
    parameter_name_prefix: T.Optional[str] = None,
    env_name: T.Optional[str] = None,
    s3folder_config: T.Optional[str] = None,
) -> Config:
    """
    获取这个 ``Config`` 对象的数据. 详细的数据结构请参考
    :class:`acore_server_config.config.main.Config`.

    :param bsm: BotoSesManager 实例.
    :param parameter_name_prefix: the parameter name prefix, the full name will
        be ${parameter_name_prefix}-${env_name}.
    :param env_name: the environment name of the env specific config you want to
        load from, stx, tst, prd, etc. If None, then load the master config.
    :param s3folder_config: S3 配置数据的根目录, 默认为
        s3://aws_account_id}-{aws_region}-artifacts/projects/acore_server_config/config/
    """
    if parameter_name_prefix is None:
        parameter_name_prefix = _get_default_parameter_name_prefix()
    if env_name is None:  # pragma: no cover
        parameter_name = parameter_name_prefix
    else:
        parameter_name = f"{parameter_name_prefix}-{env_name}"

    if s3folder_config is None:
        s3folder_config = _get_default_s3folder_config(bsm=bsm)

    config = Config.read(
        env_class=Env,
        env_enum_class=EnvEnum,
        bsm=bsm,
        parameter_name=parameter_name,
        s3folder_config=s3folder_config,
    )

    return config


@dataclasses.dataclass
class Ec2ConfigLoader:
    """
    用于在 EC2 上运行脚本, 用 "自省" 的方式获得自己的配置数据. 开始时请使用
    :meth:`Ec2ConfigLoader.load` 方法获得当前 EC2 的配置数据.

    用法:

    .. code-block:: python

        >>> server = Ec2ConfigLoader.load(...)
        >>> server
        Server(id='sbx-blue', db_admin_password='sbx*dummy4test', db_username='myuser', db_password='sbx*dummy4test')
    """

    @classmethod
    def load(
        cls,
        parameter_name_prefix: T.Optional[str] = None,
        s3folder_config: T.Optional[str] = None,
        server_id: T.Optional[str] = None,
        bsm: "BotoSesManager" = default_bsm,
    ) -> Server:
        """
        获得当前 EC2 的配置数据, 返回一个
        :class:`~acore_server_config.config.define.server.Server` 对象.

        :param parameter_name_prefix: the parameter name prefix, the full name will
            be ${parameter_name_prefix}-${env_name}.
        :param s3folder_config: S3 配置数据的根目录, 默认为
            s3://aws_account_id}-{aws_region}-artifacts/projects/acore_server_config/config/
        :param server_id: 强制指定 server_id, 跳过 "自省" 阶段. 常用于测试. 这个 server_id
            的格式为: ${env_name}-${server_name}, 例如: sbx-blue
        :param bsm: ``boto_session_manager.BotoSesManager`` object, if not provided,
            then use the current runtime default AWS CLI profile.
        """
        if server_id is None:  # pragma: no cover
            server_id = get_this_server_id(bsm=bsm)
        env_name, server_name = parse_server_id(server_id=server_id)
        config = get_config(
            bsm=bsm,
            parameter_name_prefix=parameter_name_prefix,
            env_name=env_name,
            s3folder_config=s3folder_config,
        )
        env = config.get_env(env_name)
        return env.servers[server_name]


@dataclasses.dataclass
class ConfigLoader:
    """
    用于在任意其他环境显式的加载配置数据. 开始时请使用 :meth:`ConfigLoader.new` 方法创建
    一个新的 Loader 对象, 将配置数据加载到内存中. 然后再对特定的 Server 的配置数据进行访问.

    用法:

    .. code-block:: python

        >>> config_loader = ConfigLoader.new(...)
        >>> for server_name, server in config_loader.iter_servers():
        ...
        >>> server = config_loader.get_server(server_name="blue")
        >>> server
        Server(id='sbx-blue', db_admin_password='sbx*dummy4test', db_username='myuser', db_password='sbx*dummy4test')
    """

    _env: Env = dataclasses.field(init=False)  # a cache of the env specific config

    @classmethod
    def new(
        cls,
        env_name: str,
        parameter_name_prefix: T.Optional[str] = None,
        s3folder_config: T.Optional[str] = None,
        bsm: "BotoSesManager" = default_bsm,
    ) -> "ConfigLoader":
        """
        创建一个新的 ConfigLoader 对象,

        :param env_name: the environment name of the env specific config you want to
            load from, stx, tst, prd, etc. If None, then load the master config.
        :param parameter_name_prefix: the parameter name prefix, the full name will
            be ${parameter_name_prefix}-${env_name}.
        :param s3folder_config: S3 配置数据的根目录, 默认为
            s3://aws_account_id}-{aws_region}-artifacts/projects/acore_server_config/config/
        :param bsm: ``boto_session_manager.BotoSesManager`` object, if not provided,
            then use the current runtime default AWS CLI profile.
        """
        config = get_config(
            bsm=bsm,
            parameter_name_prefix=parameter_name_prefix,
            env_name=env_name,
            s3folder_config=s3folder_config,
        )
        env = config.get_env(env_name)
        config_loader = cls()
        config_loader._env = env
        return config_loader

    def iter_servers(self) -> T.Iterable[T.Tuple[str, Server]]:
        """
        遍历所有的 server. 返回许多 (server_name, server) 的 tuple. 这类似于字典中的
        ``dict.items()`` 方法
        """
        return self._env.servers.items()

    def get_server(self, server_name: str) -> Server:
        """
        获得特定 server 的配置数据.

        :param server_name: 服务器的名字 (不包括环境名, 包括环境名的字符串是 server_id).
            例如 "blue", "green" 等.
        """
        return self._env.servers[server_name]



# -*- coding: utf-8 -*-

"""
todo: doc string
"""

import typing as T
import dataclasses

if T.TYPE_CHECKING:  # pragma: no cover
    from .main import Env


@dataclasses.dataclass
class Server:
    """
    Per Game Server configuration.

    :param id: Server id, the naming convention is ``${env_name}-${server_name}``
    :param db_admin_password: the RDS admin password, we need this password
        to create the database user.
    :param db_username: the database user for game server
    :param db_password: the database password for game server
    :param authserver_conf: custom config for authserver.conf
    :param worldserver_conf: custom config for worldserver.conf
    :param mod_lua_engine_conf: custom config for mod_LuaEngine.conf
    """

    id: T.Optional[str] = dataclasses.field(default=None)
    db_admin_password: T.Optional[str] = dataclasses.field(default=None)
    db_username: T.Optional[str] = dataclasses.field(default=None)
    db_password: T.Optional[str] = dataclasses.field(default=None)
    authserver_conf: T.Dict[str, str] = dataclasses.field(default_factory=dict)
    worldserver_conf: T.Dict[str, str] = dataclasses.field(default_factory=dict)
    mod_lua_engine_conf: T.Dict[str, str] = dataclasses.field(default_factory=dict)


@dataclasses.dataclass
class ServerMixin:
    servers: T.Dict[str, Server] = dataclasses.field(default_factory=dict)

    @property
    def server_blue(self) -> Server:
        return self.servers["blue"]

    @property
    def server_green(self) -> Server:
        return self.servers["green"]



# -*- coding: utf-8 -*-

"""
**"Runtime" Definition**

Runtime is where you execute your code. For example, if this code is running
in a CI build environment, then the runtime is "ci". If this code is running
on your local laptop, then the runtime is "local". If this code is running on
AWS Lambda, then the runtime is "lambda"

This module automatically detect what is the current runtime.

.. note::

    This module is "ZERO-DEPENDENCY".
"""

import os
import enum

IS_LOCAL = False
IS_GITHUB_CI = False
IS_EC2 = False
IS_CODEBUILD_CI = False
IS_LAMBDA = False
IS_BATCH = False
IS_FARGATE = False


class RunTimeEnum(str, enum.Enum):
    local = "loc"
    github_ci = "github_ci"
    ec2 = "ec2"
    codebuild_ci = "codebuild_ci"
    awslambda = "awslambda"
    batch = "batch"
    fargate = "fargate"
    unknown = "unknown"


CURRENT_RUNTIME: str = RunTimeEnum.unknown.value

if os.environ.get("HOME", "NA") == "/home/ubuntu":  # pragma: no cover
    IS_EC2 = True
    CURRENT_RUNTIME = RunTimeEnum.ec2.value
# if you use AWS CodeBuild for CI/CD
# ref: https://docs.aws.amazon.com/codebuild/latest/userguide/build-env-ref-env-vars.html
elif "CODEBUILD_CI" in os.environ:  # pragma: no cover
    IS_CODEBUILD_CI = True
    CURRENT_RUNTIME = RunTimeEnum.codebuild_ci.value
# if you use GitHub CI for CI/CD
# ref: https://docs.github.com/en/actions/learn-github-actions/variables
elif "CI" in os.environ:  # pragma: no cover
    IS_GITHUB_CI = True
    CURRENT_RUNTIME = RunTimeEnum.github_ci.value
# ref: https://docs.aws.amazon.com/lambda/latest/dg/configuration-envvars.html
elif "AWS_LAMBDA_FUNCTION_NAME" in os.environ:  # pragma: no cover
    IS_LAMBDA = True
    CURRENT_RUNTIME = RunTimeEnum.awslambda.value
# ref: https://docs.aws.amazon.com/batch/latest/userguide/job_env_vars.html
elif "AWS_BATCH_JOB_ID" in os.environ:  # pragma: no cover
    IS_BATCH = True
    CURRENT_RUNTIME = RunTimeEnum.batch.value
# ref: https://docs.aws.amazon.com/AmazonECS/latest/userguide/task-metadata-endpoint-v4-fargate.html
elif "ECS_CONTAINER_METADATA_URI_V4" in os.environ or "ECS_CONTAINER_METADATA_URI" in os.environ:  # pragma: no cover
    IS_FARGATE = True
    CURRENT_RUNTIME = RunTimeEnum.fargate.value
else:  # pragma: no cover
    IS_LOCAL = True
    CURRENT_RUNTIME = RunTimeEnum.local.value



# -*- coding: utf-8 -*-

from ..paths import dir_unit_test, path_config_json
from ..config.define import Env, EnvEnum, Config

path_config_secret_json = dir_unit_test.joinpath("config", "config-secret.json")

config = Config.read(
    env_class=Env,
    env_enum_class=EnvEnum,
    path_config=path_config_json,
    path_secret_config=path_config_secret_json,
)



# -*- coding: utf-8 -*-

import dataclasses
import os

from config_patterns.patterns.multi_env_json.api import (
    BaseEnvEnum,
    BaseEnv,
    BaseConfig,
)

from ...runtime import IS_LOCAL, IS_GITHUB_CI, IS_EC2, IS_CODEBUILD_CI
from ...compat import cached_property


class EnvEnum(BaseEnvEnum):
    """
    In this project, we have three environment:

    - sbx: represent the developer's local laptop, and the sandbox environment
        the change you made on your local laptop can be applied to sandbox.
    - tst: a long living integration test environment. before releasing to
        production, the app has to be deployed to test environment for QA.
    - prd: the production environment. can only be deployed from release branch.
    """

    sbx = "sbx"
    tst = "tst"
    prd = "prd"


# You may have a long list of config field definition
# put them in different module and use Mixin class
from .server import Server, ServerMixin


@dataclasses.dataclass
class Env(
    BaseEnv,
    ServerMixin,
):
    @classmethod
    def from_dict(cls, data: dict):
        data["servers"] = {
            name: Server(
                id="{}-{}".format(
                    data["env_name"],
                    name,
                ),
                **dct,
            )
            for name, dct in data.get("servers", {}).items()
        }
        env = cls(**data)
        return env


class Config(BaseConfig):
    @classmethod
    def get_current_env(cls) -> str:  # pragma: no cover
        # you can uncomment this line to force to use certain env
        # from your local laptop to run application code, tests, ...
        # return EnvEnum.sbx.value
        if IS_LOCAL:
            if "USER_ENV_NAME" in os.environ:
                return os.environ["USER_ENV_NAME"]
            return EnvEnum.sbx.value
        elif IS_GITHUB_CI or IS_CODEBUILD_CI:
            if "USER_ENV_NAME" in os.environ:
                return os.environ["USER_ENV_NAME"]
            return EnvEnum.sbx.value
        elif IS_EC2:
            if "USER_ENV_NAME" in os.environ:
                return os.environ["USER_ENV_NAME"]
            return EnvEnum.sbx.value
        else:
            raise NotImplementedError

    @cached_property
    def sbx(self) -> Env:  # pragma: no cover
        return self.get_env(env_name=EnvEnum.sbx)

    @cached_property
    def tst(self) -> Env:  # pragma: no cover
        return self.get_env(env_name=EnvEnum.tst)

    @cached_property
    def prd(self) -> Env:  # pragma: no cover
        return self.get_env(env_name=EnvEnum.prd)

    @cached_property
    def env(self) -> Env:
        return self.get_env(env_name=self.get_current_env())



# -*- coding: utf-8 -*-

from .runtime import IS_LOCAL
from .runtime import IS_GITHUB_CI
from .runtime import IS_EC2
from .runtime import IS_CODEBUILD_CI
from .config.define import EnvEnum
from .config.define import Env
from .config.define import Config
from .config.define import Server
from .config.loader import Ec2ConfigLoader
from .config.loader import ConfigLoader
from .in_ec2 import get_server

try:
    from .boto_ses import bsm
except ImportError as e:  # pragma: no cover
    pass
